/*
 * include/asm-mips/bitops.h
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (c) 1994 by Ralf Baechle
 */

#ifndef _ASM_MIPS_BITOPS_H_
#define _ASM_MIPS_BITOPS_H_

/*
 * On MIPS inline assembler bitfunctions are as effective
 * as the standard C counterparts.
 */
#include <asm-generic/bitops.h>

#endif /* _ASM_MIPS_BITOPS_H_ */
